"""Plotting module for PyTorch prototyping.

:author: Julian M. Kleber
"""
from typing import Type, Optional, Dict, Any, List, Tuple
import logging
import numpy as np
import matplotlib.pyplot as plt
from amarium.utils import load_json_from_file, prepare_file_name_saving, attach_slash

from carate.statistics.analysis import (
    get_avg,
    get_min,
    get_max,
    unpack_run,
    get_stacked_list,
    load_result_json,
    get_min_max_avg_cv_run,
)


logger = logging.getLogger(__name__)


def plot_trajectory_single_algorithm(
    path_to_directory: str,
    parameter: str,
    save_dir: str = "./plots",
    data_name: Optional[str] = None,
) -> None:
    """The plot_classification_algorithm function takes in a path to a
    directory containing the results of a classification algorithm and plots
    the accuracy of that algorithm on both training and testing data.

    :param path_to_directory: str: Used to specify the directory where
        the results are stored.
    :return: None.  :doc-author: Julian M. Kleber
    """

    path_to_directory = attach_slash(path_to_directory) + "data/"
    legend_text = path_to_directory.split("/")[-3]

    if data_name is None:
        data_name = f"{legend_text}.json"

    result = get_stacked_list(
        path_to_directory=path_to_directory,
        num_cv=5,
        json_name=data_name,
    )

    plot_range_band_single(
        result,
        file_name=f"{legend_text}_{parameter}",
        save_dir=save_dir,
        key_val=parameter,
        alpha=0.4,
        legend_text=legend_text,
    )


def plot_range_band_multi(
    result: List[Dict[str, float]],
    key_vals: List[str],
    file_name: str,
    alpha: float = 0.5,
    y_lim: Tuple[float] = (0.0, 1.01), 
    set_ylim_dynamically:bool=False,
    save_dir: Optional[str] = None,
    title_text: Optional[str] = None,
) -> None:
    """The plot_range_band_multi function is used to plot multiple range bands
    on the same graph. The function takes in a list of dictionaries, each
    dictionary containing the results from one cross-validation run.It also
    takes in a list of keys that correspond to values within each dictionary
    that should be plotted as range bands. The function then plots all these
    values as separate lines with their corresponding ranges filled in between
    them.

    :param result: List[Dict[str: Used to Pass in the list of
        dictionaries :param float]]: Used to Set the alpha value of the
        fill.
    :param key_vals: List[str]: Used to Specify which metrics to plot.
    :param file_name: str: Used to name the file that will be saved.
    :param alpha: float=0.5: Used to Set the transparency of the fill
        between.
    :param save_dir: Optional[str]=None: Used to Save the plot in a
        specific directory.
    :param : Used to Set the transparency of the fill between min and
        max values.
    :return: A plot of the average, maximum and minimum values. :doc-
        author: Julian M. Kleber
    """

    fig, axis = plt.subplots()
    axis.set_xlabel("Training step")

    for i in range(len(key_vals)):
        max_val: List[float]
        min_val: List[float]
        avg_val: List[float]

        max_val, min_val, avg_val = get_min_max_avg_cv_run(
            result=result, key_val=key_vals[i]
        )
        if set_ylim_dynamically:
            if i == 0: 
                y_min = np.min(min_val)
                y_max = np.max(max_val)
            else: 
                new_y_min = np.min(min_val)
                new_y_max = np.max(max_val)
                if y_min > new_y_min: 
                    y_min = new_y_min
                if y_max < new_y_max: 
                    y_max = new_y_max

        axis.plot(avg_val, label=key_vals[i])
        plot_range_fill(max_val, min_val, alpha, axis)

    if set_ylim_dynamically:
        y_min = y_min - 0.03*y_min
        y_max = y_max + 0.03*y_max
        axis.set_ylim(y_min, y_max)
    else: 
        axis.set_ylim(*y_lim)
    axis.set_ylabel("Value")
    axis.legend()
    axis.set_title(title_text)
    save_publication_graphic(fig_object=fig, file_name=file_name, prefix=save_dir)


def plot_range_band_single(
    result: List[Dict[str, List[float]]],
    key_val: str,
    file_name: str,
    alpha: float = 0.5,
    fixed_y_lim=(0.0, 1.01),
    set_ylim_dynamically:bool=False,
    save_dir: Optional[str] = None,
    legend_text: Optional[str] = None,
) -> None:
    """The plot_range_band function takes in a list of dictionaries, each
    dictionary containing the results from one run. It then plots the average
    value for each key_val (e.g., 'accuracy') and also plots a range band
    between the minimum and maximum values for that key_val across all runs.

    :param result: List[Dict[str: Used to plot the results of each run.
        :param float]]: Used to specify the type of data that is being
        passed into the function.
    :param key_val: str: Used to specify which key in the dictionary to
        plot.
    :param file_name: str: Used to specify the file_name on where to save the plot
    :param: alpha: float: Set the alpha level of the range band 
    :param: fixed_y_lim: str: Setting a fixed limit on the y_axis
    :param: set_ylim_dynamically: bool: If the y_lim should be set based on min, max values of 
            the defined key_val
    :param: save_dir: str: Where to save the plot 
    :param: legend_text: Text to put on the legend
    :return: None, but saves a plot with the average value of a list, and the minimum
        and maximum values. 
    :doc-author: Julian M. Kleber
    :author: Julian M. Kleber
    """
    max_val: List[float]
    min_val: List[float]
    avg_val: List[float]

    max_val, min_val, avg_val = get_min_max_avg_cv_run(result=result, key_val=key_val)

    fig, axis = plt.subplots()
    if legend_text is not None:
        axis.plot(avg_val, "-", label=legend_text)
    else:
        axis.plot(avg_val, "-", label=legend_text)
    plot_range_fill(max_val, min_val, alpha, axis)

    if set_ylim_dynamically:
        y_min = np.min(min_val) - 0.03*np.min(min_val)
        y_max = np.max(max_val) + 0.03*np.max(max_val)
        axis.set_ylim(y_min, y_max) 
    else: 
        axis.set_ylim(*fixed_y_lim)
    axis.set_ylabel(key_val)
    if legend_text is not None:
        axis.legend()

    axis.set_xlabel("Training step")

    save_publication_graphic(fig_object=fig, file_name=file_name, prefix=save_dir)


def plot_range_fill(
    max_val: List[float], min_val: List[float], alpha: float, axis
) -> None:
    """The plot_range_lines function takes in three lists of floats, max_val,
    min_val and avg_val. It then plots the average value as a line graph with
    the training steps on the x-axis and the average values on the y-axis. It
    also fills in between each point with a color to show the range of values
    for that particular step.

    :param max_val: List[float]: Used to Plot the maximum value of each
        training step.
    :param min_val: List[float]: Used to Plot the minimum value of each
        metric.
    :param avg_val: List[float]: Used to Plot the average value of a
        given metric.
    :return: A plot with the average value, max value and min values for
        each training step. :doc-author: Julian M. Kleber
    """

    training_steps = np.arange(0, len(max_val), 1)
    axis.fill_between(training_steps, min_val, max_val, alpha=alpha)


def save_publication_graphic(
    fig_object: Type[plt.figure], file_name: str, prefix: Optional[str] = None
) -> None:
    """The save_publication_graphic function saves the current figure to a
    file.

    The save_publication_graphic function saves the current figure to a
    file, with a default resolution of 300 dpi. The function also
    tightens up the layout of the plot before saving it, so that there
    is no wasted space around it in its saved form.

    :param file_name: str: Used to Specify the name of the file to be
        saved.
    :param prefix: Optional[str]=None: Used to Specify the directory
        where the file is saved.
    :return: None.  :doc-author: Julian M. Kleber
    """

    file_name = prepare_file_name_saving(
        file_name=file_name, prefix=prefix, suffix=".png"
    )
    plt.tight_layout()
    plt.savefig(file_name, dpi=300)
    logging.info("Saved plot" + str(file_name))


def parse_old_file_format_for_plot(
    path_to_json: str,
) -> List[List[float]]:  # pragma no-cover
    """
    The parse_old_file_format_for_plot function takes in a path to a json file and returns the
    following:

    :param path_to_json:str: Used to Specify the path to the json file that we want to parse.
    :return: A list of dictionaries.

    :doc-author: Julian M. Kleber
    """

    result_acc = load_result_json(path_to_json=path_to_json)
    train_frames_acc = parse_min_max_avg(result_acc["Train_acc"])
    test_frames_acc = parse_min_max_avg(result_acc["Test_acc"])
    return train_frames_acc, test_frames_acc


def load_result_json_old_format(path_to_json: str) -> Dict[str, Any]:
    """The load_result_json function takes in a path to a json file and returns
    the contents of that json file as a dictionary.The function also parses the
    "Acc" key in the dictionary, which contains lists of tuples containing
    train and test accuracy values. The function then separates these tuples
    into two separate lists, one for train accuracy values and one for test
    accuracy values.These new lists are added to the original dictionary under
    keys "Acc_train" and "Acc_val", respectively.

    :param path_to_json: str: Used to Specify the path to the json file.
    :return: A dictionary with the following keys:. :doc-author: Julian
        M. Kleber
    """

    result = load_json_from_file(path_to_json)
    acc_store = result["Acc"]
    train_acc = []
    test_acc = []

    for i in range(len(acc_store)):
        train, test = acc_store[i]
        train_acc.append(train)
        test_acc.append(test)

    del result["Acc"]
    result["Acc_train"] = train_acc
    result["Acc_val"] = test_acc

    return result


def parse_min_max_avg(result_list: List[List[float]]) -> List[float]:
    """The parse_min function takes a list of lists and returns the minimum
    value for each sublist.

    :param result_list: List[List[float]]: Used to Store the results of
        the simulation.
    :return: A list of the minimum values for each step. :doc-author:
        Julian M. Kleber
    """

    minima = []
    maxima = []
    averages = []
    for i in range(len(result_list[0])):
        step_list = result_list[:, i]
        minimum = get_min(step_list=step_list)
        maximum = get_max(step_list=step_list)
        average = get_avg(step_list=step_list)
        minima.append(minimum)
        maxima.append(maximum)
        averages.append(average)
    return [minima, maxima, averages]



"""
Evaulation object for classification
"""
import torch
import numpy as np
import numpy.typing as npt
from typing import Type, Any, Tuple, Dict, Optional
import datetime

from carate.evaluation.base import Evaluation
from carate.loader.load_data import DatasetObject
from carate.utils.model_files import save_model_parameters
from carate.models.base_model import Model


import logging



class RegressionEvaluation(Evaluation):
    """
    Module that implements the Regression evaluation
    """

    def __init__(
        self,
        dataset_name: str,
        dataset_save_path: str,
        result_save_dir: str,
        model_net: Model,
        optimizer: torch.optim.Optimizer,
        data_set: DatasetObject,
        device: torch.device,
        logger: Any, 
        resume: bool,
        test_ratio: int,
        num_epoch: int = 150,
        num_cv: int = 5,
        num_classes: int = 2,
        out_dir: str = r"./out",
        batch_size: int = 64,
        shuffle: bool = True,
        model_save_freq: int = 100,
        override: bool = True,
        normalize: bool = False,
        custom_size: Optional[int] = None,
        
    ):
        """
        The __init__ function is called when the class is instantiated.
        It sets up all the parameters needed for training and testing.

        :param self: Used to Refer to the current object.
        :param dataset_name:str: Used to Name the dataset.
        :param dataset_save_path:str: Used to Save the dataset object
         to a file.
        :param result_save_dir:str: Used to Save the results of the
        cross validation.
        :param model_net:Model: Used to Specify the model architecture.
        :param optimizer:torch.optim.Optimizer: Used to Define the optimizer
        used for training.
        :param data_set:DatasetObject: Used to Pass the dataset object
        to the class.
        :param test_ratio:int: Used to Determine the ratio of test data to
        training data.
        :param num_epoch:int=150: Used to Set the number of epochs to train for.
        :param num_cv:int=5: Used to Set the number of cross-validation folds.
        :param num_classes:int=2: Used to Set the number of classes in the dataset.
        :param out_dir:str=r"./out": Used to Define the directory where
        all results are saved.
        :param batch_size:int=64: Used to Set the batch size of the training data.
        :param shuffle:bool=True: Used to Shuffle the data set before splitting
        it into training and test sets.
        :param model_save_freq:int=100: Used to Save the model every 100 epochs.
        :param override:bool=True: Used to Override the results if they already exist.
        :param : Used to Set the number of epochs for training.
        :return: The object itself, which is then assigned to the variable "model_trainer".


        :doc-author: Julian M. Kleber
        """
        self.dataset_name = dataset_name
        self.dataset_save_path = dataset_save_path
        self.test_ratio = test_ratio
        self.num_epoch = num_epoch
        self.model_net = model_net
        self.optimizer = optimizer
        self.num_classes = num_classes
        self.num_cv = num_cv
        self.out_dir = out_dir
        self.data_set = data_set
        self.logger = logger
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.device = device
        self.result_save_dir = result_save_dir
        self.model_save_freq = model_save_freq
        self.override = override
        self.resume = resume
        self.normalize = normalize is not None
        self.custom_size = custom_size
        

        


    def cv(
        self,
        num_cv: int,
        num_epoch: int,
        num_classes: int,
        dataset_name: str,
        dataset_save_path: str,
        test_ratio: int,
        resume: bool,
        data_set: DatasetObject,
        logger:Any,
        shuffle: bool,
        batch_size: int,
        model_net: Model,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        result_save_dir: str,
        model_save_freq: int,
        override: bool = True,
        normalize: bool = True,
        custom_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        # initialize
        (
            num_cv,
            num_epoch,
            num_classes,
            dataset_name,
            dataset_save_path,
            test_ratio,
            resume,
            data_set,
            logger,
            shuffle,
            batch_size,
            model_net,
            optimizer,
            device,
            result_save_dir,
            model_save_freq,
            override,
            normalize,
            custom_size,
        ) = self._get_defaults(locals())

        # data container
        result = {}

        save_model_parameters(model_net=model_net, save_dir=result_save_dir)
        logger.logger.info("Starting " + str(num_cv) +  " CVs for "+ dataset_name +
                            " with data_stored in " + dataset_save_path )
        for i in range(num_cv):
            tmp = {}
            test_mse = []
            train_mae = []
            train_mse = []

            logger.logger.info("Starting CV "+str(i))

            loaded_dataset: torch.utils.data.Dataset
            (
                test_dataset,
                train_dataset,
                test_loader,
                train_loader,
                loaded_dataset,
            ) = data_set.load_data(
                dataset_name=dataset_name,
                dataset_save_path=dataset_save_path,
                test_ratio=test_ratio,
                batch_size=batch_size,
                shuffle=shuffle,
                custom_size=custom_size,
            )

            del train_dataset, test_dataset

            if normalize:
                norm_factor = self.__normalization_factor(
                    data_set=loaded_dataset, num_classes=num_classes
                )

            else:
                norm_factor = 1.0
            for epoch in range(1, num_epoch + 1):
                train_mae_loss, train_loss = self.train(
                    model_net=model_net,
                    epoch=epoch,
                    optimizer=optimizer,
                    train_loader=train_loader,
                    device=device,
                    norm_factor=norm_factor,
                    num_classes=num_classes,
                )
                train_mae_val, train_mse_val = self.test(
                    model_net=model_net,
                    test_loader=train_loader,
                    device=device,
                    norm_factor=norm_factor,
                    epoch=epoch,
                )
                test_mae_val, test_mse_val = self.test(
                    model_net=model_net,
                    test_loader=test_loader,
                    device=device,
                    norm_factor=norm_factor,
                    epoch=epoch,
                )

                train_mae.append(train_mae_val)
                train_mse.append(train_mse_val)
                test_mse.append(test_mae_val)
                test_mse.append(test_mse_val)

                logger.log({
                    "Epoch": epoch,
                    "Train_Loss": "{:16f}".format(train_loss),
                    "Train_MAE_Loss": "{:16f}".format(train_mae_loss),
                    "Train_MAE": "{:16f}".format(train_mae_val),
                    "Train_MSE": "{:16f}".format(train_mse_val), 
                    "Test_MAE": "{:16f}".format(test_mae_val),
                    "Test_MSE": "{:16f}".format(test_mse_val), 
                })
                torch.cuda.empty_cache()

                tmp["MAE Train"] = list(train_mae)
                tmp["MSE Train"] = list(train_mse)
                tmp["MAE Test"] = list(train_mae)
                tmp["MSE Test"] = list(train_mse)

                if epoch % model_save_freq == 0:
                    self.save_whole_checkpoint(
                        result_save_dir=result_save_dir,
                        dataset_name=dataset_name,
                        num_cv=i,
                        num_epoch=epoch,
                        model_net=model_net,
                        data=tmp,
                        optimizer=optimizer,
                        loss=train_mae_loss,
                        override=override,
                    )

            result[str(i)] = tmp
        return result

    def train(
        self,
        epoch: int,
        model_net: Model,
        device: torch.device,
        train_loader: torch.utils.data.DataLoader,
        optimizer: torch.optim.Optimizer,
        num_classes: int,
        **kwargs: Any,
    ) -> float:
        norm_factor: npt.NDArray[np.float64] = kwargs["norm_factor"]
        model_net.train()
        mse = 0
        for data in train_loader:
            data.x = data.x.type(torch.FloatTensor)

            data.y = data.y / norm_factor

            data.y = torch.nan_to_num(data.y.type(torch.FloatTensor))
            data = data.to(device)
            optimizer.zero_grad()
            output_probs = model_net(data.x, data.edge_index, data.batch)
            loss = torch.nn.MSELoss()
            if len(data.y.size()) == 1:
                loss = loss(output_probs, data.y[:])
            else:
                loss = loss(output_probs, data.y[:, :])
            mse += loss.item()
            loss.backward()
            optimizer.step()
            torch.cuda.empty_cache()
        return mse / len(train_loader), loss

    def test(
        self,
        test_loader: Type[torch.utils.data.DataLoader],
        epoch: int,
        model_net: Model,
        device: torch.device,
        **kwargs: Any,
    ) -> Tuple[float, float]:
        norm_factor: npt.NDArray[np.float64] = kwargs["norm_factor"]
        model_net.eval()
        mae = 0
        mse = 0
        for data in test_loader:
            data.x = data.x.type(torch.FloatTensor)

            data.y = data.y / norm_factor

            data.y = torch.nan_to_num(data.y.type(torch.FloatTensor))
            data = data.to(device)
            output_probs = model_net(data.x, data.edge_index, data.batch)
            loss_mae = torch.nn.L1Loss()
            mae += loss_mae(output_probs, data.y).item()
            loss = torch.nn.MSELoss()
            loss = loss(output_probs, data.y)
            mse += loss.item()
            torch.cuda.empty_cache()
        return mae / len(test_loader), mse / len(test_loader)

    def __normalization_factor(
        self, data_set: Any, num_classes: int
    ) -> npt.NDArray[np.float64]:
        """
        The __normalization_factor function is used to calculate the normalization factor
        for each class.The normalization factor is calculated by taking the L2 norm of all of the
        labels in a given class.

        This function returns an array containing one value for each class, where each value is
        equal to the L2 norm  of all labels belonging to that particular class.

        :param self: Used to Refer to the object itself.
        :param data_set:Any: Used to Pass in the data set that we want to normalize.
        :param num_classes:int: Used to Specify the number of classes in the data set.
        :return: A vector of length num_classes,.

        :doc-author: Trelent
        """

        y = np.zeros((len(data_set), 1, num_classes))
        for i in range(len(data_set)):
            y[i, :, :] = data_set[i].y
        norm_factor = np.zeros((num_classes))
        for i in range(num_classes):
            norm = np.linalg.norm(y[:, 0, i], ord=2)
            norm_factor[i] = norm
        return norm_factor

    def __repr__(self) -> str:
        return "Regression Evaluation Object"
    
    def __str(self) ->str:
        return "Regression Evaluation Object"




"""Module for routine multi-run plotting.

:author: Julian M. Kleber
"""
import os
from typing import Dict, Optional, List, Tuple
import matplotlib.pyplot as plt

from amarium.utils import attach_slash


from carate.statistics.analysis import (
    get_min_max_avg_cv_run,
    get_stacked_list,
    get_max_average,
    get_min_average
)
from carate.plotting.base_plots import plot_range_fill, save_publication_graphic

import logging

logger = logging.getLogger(__name__)



def plot_all_runs_in_dir(
    base_dir: str,
    save_name: str,
    legend_texts:List[str],
    val_single: str = "Acc_test",
    num_cv:int = 5, 
    y_lims=(0.0, 1.01),
) -> None:
    """
    Function to plot hyperparameter tunins of a single dataset and algorithm inside 
    a directory

    :author: Julian M. Kleber
    """

    run_dirs = os.listdir(base_dir)
    fig, axis = plt.subplots()
    for i in range(len(run_dirs)):
        
        result = prepare_plot_multi(base_dir=base_dir, run_dir=run_dirs[i], val_single=val_single, num_cv=num_cv)
        plot_range_band_multi_run(
            result,
            fixed_y_lim=y_lims,
            key_val=val_single,
            file_name=f"{legend_texts[i]}_{val_single}",
            save_dir="./plots",
            alpha=0.4,
            legend_text=legend_texts[i],
            fig=fig,
            axis=axis
        )
        
    save_publication_graphic(fig_object=fig, file_name=save_name)



def ploat_range_band_multi_val()->None: 


    pass

def plot_range_band_multi_run(
    result: List[Dict[str, List[float]]],
    key_val: str,
    file_name: str,
    fig,
    axis,
    alpha: float = 0.5,
    fixed_y_lim=(0.0, 1.01),
    save_dir: Optional[str] = None,
    legend_text: Optional[str] = None,
) -> None:
    """The plot_range_band function takes in a list of dictionaries, each
    dictionary containing the results from one run. The function is meant to be
    used in a for-loop iterating about many runs. It then plots the average
    value for each key_val (e.g., 'accuracy') and also plots a range band
    between the minimum and maximum values for that key_val across all runs.

    :param result: List[Dict[str: Used to plot the results of each run.
        :param float]]: Used to specify the type of data that is being
        passed into the function.
    :param key_val: str: Used to specify which key in the dictionary to
        plot.
    :param file_name: str: Used to save the plot as a png file.
    :return: A plot with the average value of a list, and the minimum
        and maximum values. :doc-author: Julian M. Kleber
    """
    max_val: List[float]
    min_val: List[float]
    avg_val: List[float]

    max_val, min_val, avg_val = get_min_max_avg_cv_run(result=result, key_val=key_val)

    if legend_text is not None:
        axis.plot(avg_val, "-", label=legend_text)
    else:
        axis.plot(avg_val, "-", label=legend_text)

    plot_range_fill(max_val, min_val, alpha, axis)

    axis.set_ylim(*fixed_y_lim)
    axis.set_ylabel(key_val)
    if legend_text is not None:
        axis.legend()

    axis.set_xlabel("Training step")


def prepare_plot_multi(base_dir:str, run_dir:str, val_single:str, num_cv:int=5):

    
    
    full_dir = attach_slash(base_dir) + attach_slash(run_dir) + attach_slash("data")

    name = os.listdir(full_dir + attach_slash("CV_0"))[0]

    logger.info("Full dir for run to plot:", full_dir)
    legend_text = full_dir.split("/")[-3]
    logger.info("Plotting: ", legend_text)
    result = get_stacked_list(
            path_to_directory=full_dir,
            num_cv=num_cv,
            json_name=name,
    )

    return result


import os

from carate.logging.metrics_logger import MetricsLogger
from amarium.utils import (
    delete_file,
    delete_all_files_in_dir,
    delete_empty_dir,
    make_full_filename,
)


def test_metrics_logger():
    if os.path.isdir("tests/test_loggers/"):
        delete_all_files_in_dir("tests/test_loggers/")

    save_dir = "tests/test_loggers/"

    metricsLogger = MetricsLogger(save_dir)
    metricsLogger.log({"MAE": 0.001}, "info")
    assert os.path.basename(metricsLogger.filename) in os.listdir("tests/test_loggers")

    delete_all_files_in_dir("tests/test_loggers/")
    delete_empty_dir("tests/test_loggers/")

    assert not os.path.isdir("test/test_loggers/")



import os
import pandas as pd
import logging

from carate.plotting.base_plots import plot_trajectory_single_algorithm
from carate.plotting.multi_run import plot_all_runs_in_dir

from tests.utils import check_plotting_dir


def test_multi_plot_regression():
    check_plotting_dir()
    path_to_directory = "./notebooks/data/Regression/"
    parameter = "MAE Train"
    save_name = "./plots/multi_run_regression.png"

    plot_all_runs_in_dir(
        base_dir=path_to_directory,
        save_name=save_name,
        val_single=parameter,
        legend_texts=[
            "ALCHEMY_20",
            "ALCHEMY_20_test_training_length_no_norm",
            "ZINC_20",
            "ALCHEMY_20_no_norm",
            "ZINC_20_no_norm",
        ],
    )

    assert os.path.isdir("./plots")
    assert os.path.isfile(save_name)


def test_multi_plot_classification():
    check_plotting_dir()
    path_to_directory = "./notebooks/data/Classification/"
    parameter = "Acc_train"
    save_name = "./plots/multi_run_classification.png"

    plot_all_runs_in_dir(
        path_to_directory,
        save_name=save_name,
        y_lims=(0.009, 0.0009),
        legend_texts=['PROTEINS', 'ENZYMES', 'MCF-7', 'YEAST', 'MOLT-4'],
        val_single=parameter,
    )

    assert os.path.isdir("./plots")
    assert os.path.isfile(save_name)


def test_classification_plot_run():
    check_plotting_dir()
    path_to_directory = "./notebooks/data/Classification/ENZYMES/"
    parameter = "Acc_test"

    plot_trajectory_single_algorithm(
        path_to_directory=path_to_directory, parameter=parameter
    )
    assert os.path.isdir("./plots")
    assert os.path.isfile(f"./plots/ENZYMES_{parameter}.png")


def test_regression_plot_run():
    check_plotting_dir()
    path_to_directory = (
        "./notebooks/data/Regression/ALCHEMY_20_test_training_length_no_norm/"
    )
    parameter = "MAE Train"
    data_name = "alchemy_full.json"

    plot_trajectory_single_algorithm(
        path_to_directory=path_to_directory, parameter=parameter, data_name=data_name
    )
    assert os.path.isdir("./plots")
    assert os.path.isfile(
        f"./plots/ALCHEMY_20_test_training_length_no_norm_{parameter}.png"
    )



"""Module providing a CLI
interface for training models#

:author: Julian  M. Kleber
"""
import os
from typing import Type
import click

from carate.runner.run import RunInitializer, Run


@click.group(invoke_without_command=True)
@click.pass_context
@click.option("-c", help="Path to config file")
@click.option("-d", help="Path to directory")
def carate(ctx, c, d):
    if ctx.invoked_subcommand is None:
        start_run(c=c, d=d)
    else:
        click.echo(f"I am about to invoke {ctx.invoked_subcommand}")


def start_run(c: str, d: str) -> None:
    """
    The train_algorithm function takes in a config file and an output directory.
    It then runs the algorithm using the configuration specified in the config file,
    and saves all of its outputs to the output directory.

    :param c:str: Used to specify the path to the configuration file.
    :param d:str: Used to specify the path to the configuration file.

    :return: None.

    :doc-author: Julian M. Kleber
    """
    if c is None and d is None:
        raise RuntimeError("Please provide the path to a config file.")
    elif c != None and d != None:
        raise RuntimeError("Please provide either a directory or a file path.")
    elif c != None and d == None:
        config_filepath = c
        runner = RunInitializer.from_file(config_filepath=config_filepath)

        runner.run()
    elif d != None and c == None:
        train_whole_directory(d)


def train_whole_directory(d: str) -> None:
    """
    The train_algorithm function takes in a config file and an output directory.
    It then runs the algorithm using the configuration specified in the config file,
    and saves all of its outputs to the output directory.

    :param c:str: Used to specify the directory to the configuration files.
    :param o:str: Used to specify the output directory.
    :return: None.

    :doc-author: Julian M. Kleber
    """

    if d is None:
        raise RuntimeError("Please provide the path to a directory of config files")

    directory_file_path = d
    config_files = [
        file
        for file in os.listdir(directory_file_path)
        if (file.endswith(".py") or file.endswith(".json") or file.endswith(".yml"))
    ]
    if len(config_files) == 0:
        raise RuntimeError(
            f"The directory {directory_file_path} does not have any config files"
        )

    for config_file in config_files:
        if config_file.endswith(".py"):
            runner = RunInitializer.from_file(config_filepath=config_file)
        elif config_file.endswith(".json"):
            runner = RunInitializer.from_json(config_filepath=config_file)
        else:
            continue

        runner.run()



"""
cgc_regression model is named after the structure of the graph neural network.
The graph neural network is structured with a convolutional , graph attention,
and another convolutional layer. The cgc_classificatin model was the model tested int the publication
Introducing CARATE: Finally speaking chemistry.
"""
import numpy as np
import torch
import torch.nn.functional as F
from torch.nn import Linear
from torch_geometric.nn import global_add_pool, GraphConv, GATConv
import sklearn.metrics as metrics

import logging

from carate.models.base_model import Model


logger = logging.getLogger(__name__)
logging.basicConfig(
    filename="carate.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


class Net(Model):
    def __init__(
        self,
        dim: int,
        num_features: int,
        num_classes: int,
        num_heads: int = 16,
        dropout_gat: float = 0.6,
        dropout_forward=0.5,
        *args,
        **kwargs,
    ) -> None:
        super(Net, self).__init__(
            dim=dim, num_classes=num_classes, num_features=num_features
        )

        self.num_heads = num_heads
        self.dropout_gat = dropout_gat
        self.dropout_forward = dropout_forward

        self.conv1 = GraphConv(self.num_features, self.dim)
        self.gat1 = GATConv(self.dim, self.dim, dropout=self.dropout_gat)
        self.conv2 = GraphConv(self.dim, self.dim)

        self.fc1 = Linear(self.dim, self.dim)
        self.fc2 = Linear(self.dim, self.num_classes)

    def forward(self, x: float, edge_index: int, batch: int, edge_weight=None) -> float:
        x = F.relu(self.conv1(x, edge_index, edge_weight))

        x = F.dropout(x, p=self.dropout_forward, training=self.training)
        x = F.relu(self.gat1(x, edge_index, edge_weight))
        x = F.relu(self.conv2(x, edge_index, edge_weight))
        x = global_add_pool(x, batch)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.fc2(x)
        return x
    def __str__(self): 
        return "cgc_classification"